#!/usr/bin/env python

from itertools import combinations
from random import randint
import EXIF
import PIL.Image
import math
import sys

corners = [
    (0, 0, 0), (255, 0, 0), (0, 255, 0), (255, 255, 0),
    (0, 0, 255), (255, 0, 255), (0, 255, 255), (255, 255, 255)
    ]

def distance(rgb1, rgb2):
    r1, g1, b1 = rgb1
    r2, g2, b2 = rgb2
    return math.sqrt(pow(r1 - r2, 2) + pow(g1 - g2, 2) + pow(b1 - b2, 2))

def brightness(rgb):
    return distance(rgb, (0, 0, 0))

def factor(start, end, rgb):
    r1, g1, b1 = start
    r2, g2, b2 = end
    gr, gg, gb = r2 - r1, g2 - g1, b2 - b1
    r, g, b = rgb
    pr, pg, pb = r - r1, g - g1, b - b1
    dp = pr * gr + pg * gg + pb * gb
    return dp / pow(distance(start, end), 2)

def darklight(rgb1, rgb2):
    if brightness(rgb1) <= brightness(rgb2):
        return rgb1, rgb2
    else:
        return rgb2, rgb1

def nearest(rgb, values=None):
    l = map(lambda c: (distance(rgb, c), c), values or corners)
    l.sort()
    return l

def pattern(rgb):
    l = nearest(rgb)
    start, end = l[0][1], l[1][1]
    f = factor(start, end, rgb)
    #if f > 0.5:
    #    start, end = end, start
    #    f = 1 - f
    return start, end, f

def choose(seq, f):
    last = int(seq * math.sqrt(f))
    current = int((seq + 1) * math.sqrt(f))
    return last != current

def get_value(xy, rgb, width, height):
    x, y = xy
    rgb1, rgb2, f = pattern(rgb)
    if choose(x + randint(0, width), f) and choose(y + randint(0, height), f):
        return rgb2
    else:
        return rgb1

def get_best(rgb, values):
    return nearest(rgb, values)[0][1]

def sign(x):
    return x >= 0 and 1 or -1

def saturate(rgb):
    return tuple(map(saturate_value, rgb))

def saturate_value(x):
    return 127.5 + sign(x - 127.5) * 127.5 * math.sqrt(abs(x - 127.5) / 127.5)

def test():
    size = 512
    for r in (0, 63, 127, 191, 255):
        im = PIL.Image.new("RGB", (size, size))
        for g in range(0, size):
            for b in range(0, size):
                value = get_value((g, b), (r, (g * 256) / size, (b * 256 / size)), size, size)
                im.putpixel((g, b), value)
        im.save("rgb%d.png" % r)

def test_flat(rgb):
    size = 64
    im = PIL.Image.new("RGB", (size, size))
    for y in range(0, size):
        for x in range(0, size):
            im.putpixel((x, y), get_value((x, y), rgb, size, size))
    im.save("rgb%02d%02d%02d.png" % rgb)

def rotate_and_scale(im, width, height, rotate):
    if rotate or x and x["Image Orientation"].values == [6L]:
        im = im.rotate(270)

    w, h = im.size
    if w > h:
        height = (width * h) / w
    else:
        width = (height * w) / h

    return im.resize((width, height))

if __name__ == "__main__":
    if "--test" in sys.argv:
        test()
        sys.exit(0)
    elif "--test-flat" in sys.argv:
        test_flat((120, 40, 60))
        sys.exit(0)

    width = 320
    height = 256

    input_filename, output_filename = sys.argv[1:3]
    rotate = "-r" in sys.argv[3:]
    saturate = "-s" in sys.argv[3:]

    x = EXIF.process_file(open(input_filename))
    im = PIL.Image.open(input_filename)
    im = rotate_and_scale(im, width, height, rotate)

    width, height = im.size

    colours = []

    for y in range(0, height):
        c = {}
        for x in range(0, width):
            rgb = im.getpixel((x, y))
            if saturate:
                rgb = saturate(rgb)
            value = get_value((x, y), rgb, width, height)
            im.putpixel((x, y), value)
            if not c.has_key(value):
                c[value] = 1
            else:
                c[value] += 1
        c = [(n, value) for value, n in c.items()]
        c.sort(reverse=True)
        colours.append(c)

    for y, c in enumerate(colours):
        if len(c) <= 4:
            continue
        most = [value for n, value in c[:4]]
        least = [value for n, value in c[4:]]

        for x in range(0, width):
            rgb = im.getpixel((x, y))
            if rgb in least:
                value = get_best(rgb, most)
                im.putpixel((x, y), value)

    im.save(output_filename)

# vim: tabstop=4 expandtab shiftwidth=4
